from masserstein import peptides
from string import printable
from collections import Counter

for i in range(255):
    if chr(i) in printable and not chr(i) in "\t\n" and not i in [11, 12, 13]:
        s = ["/* Code:", str(i).rjust(3), " ASCII char:", chr(i), "*/      "]
    else:
        s = ["/* Code:", str(i).rjust(3), " unprintable   */      "]

    chnosse = list("CHNOS") + ["Se"]
    s.append(", ".join(str(peptides.daminoacids.get(chr(i).upper(), Counter())[el]) for el in chnosse)+',')
    print(' '.join(s))

    



from tqdm import tqdm

import cppyy
import glob

for header in glob.glob("../../IsoSpec++/*.h"):
    if not "mman.h" in header:
        cppyy.include(header)
cppyy.load_library("../../IsoSpec++/libIsoSpec++.so")


t = 0.0
for x in tqdm(xrange(1000000)):
    i = cppyy.gbl.IsoSpec.Iso("C100H100N100O100")
    t += i.getTheoreticalAverageMass()
print t



from cffi import FFI

ffibuilder = FFI()
# Note that the actual source is None
ffibuilder.set_source("_simple_example", None)
ffibuilder.cdef("""
    int printf(const char *format, ...);
""")

if __name__ == "__main__":
    ffibuilder.compile(verbose=True)



from IsoSpecPy.IsoSpecPy import IsoTotalProb

# isotopic distribution of protein with fasta sequence AAAPPGQAAC
MZI_opt = IsoTotalProb(.999,
                       fasta='AAAPPGQAAC')

# isotopic distribution of protein with fasta sequence AAAPPGQAAC,
# with thioglycinie modification
MZI_opt2 = IsoTotalProb(.999,
                        formula='O-1S1',
                        fasta='AAAPPGQAAC')
print("Masses and probabilities in AAAPPGQAAC with thioglycine modification:")
print([(m, p) for (m, p) in zip(MZI_opt2.masses, MZI_opt2.probs)])




import cppyy
import glob
import re

for header in glob.glob("../../../IsoSpec++/*.h"):
    if not "mman.h" in header:
        cppyy.include(header)
cppyy.load_library("../../../IsoSpec++/libIsoSpec++.so")


regex_pattern = re.compile('([A-Z][a-z]?)([0-9]*)')

def IsoParamsFromFormula(formula):
    global regex_pattern

    symbols = []
    atomCounts = []
    for elem, cnt in re.findall(regex_pattern, formula):
        symbols.append(elem)
        atomCounts.append(int(cnt) if cnt is not '' else 1)
    try:
        masses = tuple(PeriodicTbl.symbol_to_masses[s] for s in symbols)
        probs  = tuple(PeriodicTbl.symbol_to_probs[s]  for s in symbols)
    except KeyError:
        raise ValueError("Invalid formula")

    return (atomCounts, masses, probs)





def binom(n, k):
    """Quickly adapted from https://stackoverflow.com/questions/26560726/python-binomial-coefficient"""
    if k < 0 or k > n:
        return 0
    if k == 0 or k == n:
        return 1
    total_ways = 1
    for i in range(min(k, n - k)):
        total_ways = total_ways * (n - i) // (i + 1)
    return total_ways


def max_confs_cnt(formula=""):
    """Get the maximal number of configurations for a given chemical formula."""
    from IsoSpecPy import IsoParamsFromFormula
    f = IsoParamsFromFormula(formula)
    if f.atomCount:
        N = 1
        for n, p in zip(f.atomCount, f.prob):
            N *= binom(n+len(p)-1, n) 
        return N
    else:
        return 0


def test_max_confs_cnt():
    assert max_confs_cnt("O100") == 5151
    assert max_confs_cnt("O100N10S6") == 4759524


test_formulas = [   'O100',
                    'O100N10S6',
                    'C100H202',
                    'S10H20'        ]

def test_all_configs_output_cnt():
    """Test if IsoSpecPy output correctly all configurations."""
    from IsoSpecPy import IsoThreshold
    global test_formulas
    for f in test_formulas:
        I = IsoThreshold(formula=f, threshold=0.0, absolute=True)
        assert len(I) == max_confs_cnt(f)
    print("Seems OK!")

if __name__ == "__main__":
    test_all_configs_output_cnt()



# Calculates the isotopic distribution of water in several ways

'''
We shall walk through a set of configurations which covers at least 99.9% of the total
probability. For water we could obviously go through the entire spectrum (100%), but for
larger molecules the entire spectrum has far too many configurations. Luckily, most of the
likelihood is concentrated in a relatively small set of most probable isotopologues
- and this method allows one to quickly calculate such a set, parametrising on the
percentage of coverage of total probability space required.

This is usually better than just calculating all isotopes with probabilities above a
iven threshold, as it allows one to directly parametrise on the accuracy of the
simplified spectrum - that is, the L1 distance between the full and simplified spectrum
will be less than (in this example) 0.001.

If for some reason one would wish to just calculate a set of peaks with probabilities
above a given threshold - it is possible using the IsoThresholdGenerator class.

Note: the returned set will usually contain a bit more configurations than necessary
to achieve the desired coverage. These configurations need to be computed anyway, however
it is possible to discard them using the optional trim argument.
'''

import IsoSpecPy
from math import exp

try:
    if IsoSpecPy.__version__[:4] != '2.1.':
        raise AttributeError
except AttributeError:
    print("This file is meant to be used with IsoSpecPy version 2.0.X. You seem to have a different version installed on your system.")
    import sys
    sys.exit(-1)

i = IsoSpecPy.IsoTotalProb(formula="H2O1", prob_to_cover = 0.999, get_confs=True)

print("Calculating isotopic distribution of water. Here's a list of configurations necessary to cover at least 0.999 of total probability:")

for mass, prob, conf in i:
    print("")
    print("Mass: " + str(mass))
    print("probability: " + str(prob))
    print("Number of Protium atoms: " + str(conf[0][0]))
    print("Number of Deuterium atoms: " + str(conf[0][1]))
    print("Number of O16 atoms: " + str(conf[1][0]))
    print("Number of O17 atoms: " + str(conf[1][1]))
    print("Number of O18 atoms: " + str(conf[1][2]))

print("")
print("Now what if both isotopes of hydrogen were equally probable, while prob. of O16 was 50%, O17 at 30% and O18 at 20%?")

hydrogen_probs = (0.5, 0.5)
oxygen_probs = (0.5, 0.3, 0.2)
hydrogen_masses = (1.00782503207, 2.0141017778)
oxygen_masses = (15.99491461956, 16.99913170, 17.9991610)

i = IsoSpecPy.IsoTotalProb(atomCounts = (2, 1), isotopeMasses = (hydrogen_masses, oxygen_masses), isotopeProbabilities = (hydrogen_probs, oxygen_probs), prob_to_cover = 0.999, get_confs=True)



print("The first configuration has the following parameters:")
print("Mass: " + str(i.masses[0]))
print("probability: " + str(i.probs[0]))
conf = i.confs[0]
conf_hydrogen = conf[0]
conf_oxygen = conf[1]
print("Number of Protium atoms: " + str(conf_hydrogen[0]))
print("Number of Deuterium atoms: " + str(conf_hydrogen[1]))
print("Number of O16 atoms: " + str(conf_oxygen[0]))
print("Number of O17 atoms: " + str(conf_oxygen[1]))
print("Number of O18 atoms: " + str(conf_oxygen[2]))









from .IsoSpecPy import Iso
from collections import Counter



loratadine = "C22H23N2O2Cl1"
desloratadine = "C19H19N2Cl1"
titin = "C169719H270466N45688O52238S911"
bovine_insulin = "C254H377N65O75S6"
ubiquitin = "C378H629N105O118S1"
substance_p = "C63H98N18O13S1"
protein_p16 = "C681H1100N216O208S5"
cholesterol = "C27H46O1"
caffeine = "C8H10N4O2"
water = "H2O1"
glucose = "C6H12O6"
surcose = "C12H22O11"
horse_myoglobin = "C769H1212N210O218S2"
oxygen = "O2"





averagine_unit = Counter({
'C' : 4.9384,
'H' : 7.7583,
'N' : 1.3577,
'O' : 1.4773,
'S' : 0.0417})

averagine_mass = 111.1254

def fillHs(d, target_avg_mass):
    while True:
        i = Iso(formula = d)
        if i.getTheoreticalAverageMass() > target_avg_mass:
            break
        d['H'] += 1

def averagine_dct(target_avg_mass):
    units = float(target_avg_mass) / averagine_mass
    d = dict((key, int(units*val)) for key, val in averagine_unit.items())
    fillHs(d, target_avg_mass)
    for x in list(d.keys()):
        if d[x] == 0:
            del d[x]
    return d

def averagine(target_avg_mass):
    return ''.join(str(key)+str(val) for key, val in sorted(averagine_dct(target_avg_mass).items(), key= lambda x: x[0]))


acgt_avg_base = Counter({
'C' : 9.75,
'H' : 12.25,
'N' : 3.75,
'O' : 6.0,
'P' : 1.0
})

def acgt_dct(no_bases, add_water = True):
    c = {key : int(val*no_bases+0.5) for key, val in acgt_avg_base.items()}
    if add_water:
        c['H'] += 2
        c['O'] += 1
    return c

def acgt(no_bases, add_water = True):
    c = acgt_dct(no_bases, add_water = add_water)
    return ''.join(str(key)+str(val) for key, val in sorted(c.items(), key= lambda x: x[0]))



from .IsoSpecPy import *


__version__ = "2.1.3"

# Old, deprecated name, for compatibility with 1.9.X only
IsoLayered = IsoTotalProb

# For backward compatibility with 1.0.X:
class CompatIsoWrapper(object):
    def __init__(self):
        from .IsoSpecPyOld import IsoSpec, IsoSpecify, IsoPlot
        self.IsoSpec = IsoSpec
        self.IsoSpecify = IsoSpecify
        self.IsoPlot = IsoPlot


IsoSpecPy = CompatIsoWrapper()








class ConfsPassthrough(object):
    def __init__(self, confs_parser, size):
        self.confs_parser = confs_parser
        self.size = size

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        return self.confs_parser(idx)






"""A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject
"""

# Always prefer setuptools over distutils
from setuptools import setup, find_packages, Extension
from distutils import spawn
# To use a consistent encoding
from codecs import open
import os
import glob
import shutil


# For debugging, set the below to True. Run with (sth like): LD_PRELOAD='/usr/lib64/gcc/x86_64-pc-linux-gnu/9.2.0/libasan.so.5.0.0' python ...
use_asan = False

if not os.path.exists("IsoSpec++"):
    try:
        os.symlink("../IsoSpec++", "IsoSpec++")
    except: # OS doesn't support symlinks
        shutil.copytree("../IsoSpec++", "IsoSpec++", symlinks=True)



here = os.path.abspath(os.path.dirname(__file__))

# Get the long description from the relevant file
#with open(path.join(here, 'DESCRIPTION.rst'), encoding='utf-8') as f:
#    long_description = f.read()

if use_asan:
    cmodule = Extension('IsoSpecCppPy',
                    sources = ['IsoSpec++/unity-build.cpp'],
                    extra_compile_args = '-O0 -g -DISOSPEC_DEBUG -std=c++17 -fsanitize=address'.split(),
                    extra_link_args = '-fsanitize=address'.split()
                    )
else:
    cmodule = Extension('IsoSpecCppPy',
                    sources = ['IsoSpec++/unity-build.cpp'],
                    extra_compile_args = '-mtune=native -march=native -O3 -std=c++17'.split(),
                    )


setup_args = {
#setup(
    'name': 'IsoSpecPy',

    # Versions should comply with PEP440.  For a discussion on single-sourcing
    # the version across setup.py and the project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    'version': '2.1.3',

    'description': 'Python interface to IsoSpec++ isotopic envelope calculator library',
    'long_description': 'Python interface to IsoSpec++ isotopic envelope calculator library',

    # The project's main homepage.
    'url': 'http://matteolacki.github.io/IsoSpec/',

    # Author details
    'author': 'Mateusz Lacki & Michal Startek',
    'author_email': 'matteo.lacki@gmail.com',

    # Choose your license
    'license': '2-clause BSD',

    # See https://pypi.python.org/pypi?%3Aaction=list_classifiers
    'classifiers' : [
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Chemistry',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: BSD License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],

    # What does your project relate to?
    'keywords' : 'isotopic envelope mass spectrometry',

    # You can just specify the packages manually here if your project is
    # simple. Or you can use find_packages().
    'packages' : ['IsoSpecPy'],#find_packages(),

    # List run-time dependencies here.  These will be installed by pip when
    # your project is installed. For an analysis of "install_requires" vs pip's
    # requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    'install_requires' : ['cffi'],

    'zip_safe' : False,

    # List additional groups of dependencies here (e.g. development
    # dependencies). You can install these using the following syntax,
    # for example:
    # $ pip install -e .[dev,test]
    'extras_require' : {
       'test': ["pytest", "numpy"]
    },

    # If there are data files included in your packages that need to be
    # installed, specify them here.  If using Python 2.6 or less, then these
    # have to be included in MANIFEST.in as well.
    'package_data' : {},

    # Although 'package_data' is the preferred approach, in some case you may
    # need to place data files outside of your packages. See:
    # http://docs.python.org/3.4/distutils/setupscript.html#installing-additional-files # noqa
    # In this case, 'data_file' will be installed into '<sys.prefix>/my_data'
    'data_files' : [], #[('my_data', ['data/data_file'])],

    # To provide executable scripts, use entry points in preference to the
    # "scripts" keyword. Entry points provide cross-platform support and allow
    # pip to create the appropriate form of executable for the target platform.
#    entry_points={
#        'console_scripts': [
#            'sample=sample:main',
#        ],
#    },
    'ext_modules' : [cmodule],
}

import platform
import sys


if 'windows' in platform.system().lower():
    # Probably Anaconda distribution.
    # Of course Windows can't even compile stuff. Install prebuilt C++ lib.
    import copy
    win_setup_args = copy.deepcopy(setup_args)
    win_setup_args['ext_modules'] = []
    win_setup_args['include_package_data'] = True
    setup(**win_setup_args)
elif 'cygwin' in platform.system().lower():
    try:
        import cffi
    except ImportError:
        print("You appear to be using CYGWIN, and CFFI was not found. Please use the Cygwin installer to install the cffi-python package for the appropriate Python version.")
        print("Installing CFFI using pip will most likely NOT work. This is *NOT* a bug in IsoSpecPy.")
        sys.exit(1)
    if spawn.find_executable('clang++') == None:
        print("You appear to be using CYGWIN and clang++ executable was not found. Please install the clang++ package using Cygwin installer.")
        sys.exit(1)
    import distutils
    import distutils.sysconfig
    distutils.sysconfig.get_config_vars() # Precalculate the dict so we can...
    distutils.sysconfig._config_vars['CFLAGS'] = "" # Nuke CFLAGS: they contain gcc stuff not supported by clang

    from distutils.command.build_ext import build_ext
    class build_ext_subclass(build_ext):
        def get_libraries(self, ext):
            return ext.libraries # Override default function which wants to link in libpython on Windows. We're using CFFI and don't need that.
    setup_args['cmdclass'] = {'build_ext' : build_ext_subclass}
    setup(**setup_args)

elif 'darwin' in platform.system().lower():
    # Okay, so OSX is apparently horribly broken. On OSX the "g++" command can be nonexistent and stuff will be compiled with clang++,
    # or g++ can be present and behave sanely, or it can be clang pretending to be g++ and behaving somewhat sanely, or it can be broken
    # clang which needs extra commandline arguments to compile anything that imports stdlib headers. Because hey, it's absolutely normal
    # that C++ compiler should require extra commandline args to compile a "Hello World" program, and of course those commandline options
    # are incompatible with g++ it's pretending to be. Someone at Apple should be tarred and feathered for this. For now we will do
    # some convoluted logic to try to find out experimentally whether "g++" really is g++, or is working clang or is broken clang, and 
    # which flags are needed to compile stuff. Setuptools don't make this easy to do as well...
    #
    # See https://github.com/mciach/wassersteinms/issues/1 for the rationale behind this.

    from distutils.command.build_ext import build_ext
    import subprocess

    class build_ext_subclass(build_ext):
        def build_extensions(self):
            compiler_cmd = "g++" # sane default in case the next stuff crashes...
            try:
                compiler_cmd = self.compiler.compiler_cxx[0]
                compiler_cmd = self.compiler.compiler_so_cxx[0]
            except AttributeError:
                pass

            with open(os.devnull, 'w') as devnull:

                def check_flags(flags_l):
                    cpp_hello_world = '''#include <iostream>

                    int main()
                    {
                        std::cout << "Hello World!" << std::endl;
                    };
                    '''
                    try:
                        proc = subprocess.Popen([compiler_cmd] + ["-x", "c++", "-fsyntax-only", "-"] + flags_l, stdin = subprocess.PIPE, stdout = devnull, stderr = devnull, universal_newlines = True)
                        proc.stdin.write(cpp_hello_world)
                        proc.stdin.flush()
                        proc.stdin.close()
                        ret = proc.wait()
                    except (OSError, IOError):
                        return False

                    print("Check flags:", ret == 0, flags_l)
                    return ret == 0

                if check_flags(["-stdlib=libc++"]):
                    cmodule.extra_compile_args.append("-stdlib=libc++")
                elif check_flags(["-stdlib=libstdc++"]):
                    cmodule.extra_compile_args.append("-stdlib=libstdc++")
                # else just hope for the best, that is, that the compiler isn't broken...
            try:
                super(build_ext_subclass, self).build_extensions()
            except TypeError: # which means we're on Python 2 and *of course* build_ext is an old-style class...
                build_ext.build_extensions(self)

    setup_args['cmdclass'] = {'build_ext' : build_ext_subclass}
    setup(**setup_args)

else:
    # Assuming sane UNIX with a working compiler.
    setup(**setup_args)



# Calculates the isotopic distribution of isotopically labelled glucose, where two of the carbons are replaced with 14C
# We assume that radiolabelling is 95% efficient (that is, there is a 95% chance for each of the radiolabel atoms to be 14C, and
# a 5% chance of them being either 12C or 13C, with the probability of each of those proportional to their standard isotopic abundances)

import IsoSpecPy
from IsoSpecPy import PeriodicTbl
from math import exp

try:
    if IsoSpecPy.__version__[:4] != '2.1.':
        raise AttributeError
except AttributeError:
    print("This file is meant to be used with IsoSpecPy version 2.0.X. You seem to have a different version installed on your system.")
    import sys
    sys.exit(-1)



# Now, the radiolabelled carbon
# 14C isn't normally considered in the isotopic distribution, here we add an extra isotope to the standard ones
radiolabelled_carbon_masses = PeriodicTbl.symbol_to_masses["C"] + (14.003241989,)

# Assuming that the labelling was only 95% efficient, that is only 95% 
# of the radiolabel atoms have standard C replaced with 14C. Non-replaced atoms have standard
# isotopic abundance (realtive to each other)
normal_carbon_probs = PeriodicTbl.symbol_to_probs["C"]
radiolabelled_carbon_probs = (0.05*normal_carbon_probs[0], 0.05*normal_carbon_probs[1], 0.95)

i = IsoSpecPy.IsoTotalProb(formula = "C4H12O6", # The formula for glucose, sans the radiolabel atoms
                                  # Here we specify additional "elements" which occur *in addition* to those from the formula
                                  atomCounts = (2,),
                                  isotopeMasses = (radiolabelled_carbon_masses,), 
                                  isotopeProbabilities = (radiolabelled_carbon_probs,), 
                                  # And the rest of parameters for configuration
                                  prob_to_cover = 0.99,
                                  get_confs=True)

# Radiolabelling (or isotopic labelling) with more than one element looks like this: 
# Let's say we wanted to have glucose with one 14C carbon, and two deuteriums, all with 95% probability
# Then it would be:
#i = IsoSpecPy.IsoLayeredGenerator(formula = "C5H10O6", # The formula for glucose, sans the radiolabel atoms
#                                  atomCounts = (1, 2),   
#                                  isotopeMasses = (radiolabelled_carbon_masses, PeriodicTbl.symbol_to_masses["H"]),
#                                  isotopeProbabilities = (radiolabelled_carbon_probs, (0.05, 0.95)),
#                                  # And the rest of parameters for configuration
#                                  prob_to_cover = 0.99,
#                                  get_confs=True)



print("The list of configurations that, taken together, cover at least 99% of the probability space is:")

for mass, prob, conf in i:
    print("=====================================================")

    print("Mass: " + str(mass))
    print("probability: " + str(prob))
    
    # The configuration fragments corresponding to the elements from the formula is first, the maually specified elements follow in order
    print("Number of 12C atoms: " + str(conf[0][0] + conf[3][0])) # Counting the normal and unsuccesfully radiolabelled atoms
    print("Number of 13C atoms: " + str(conf[0][1] + conf[3][1]))
    print("Number of 14C atoms: " + str(conf[3][2])) # Note: conf[0][2] will raise IndexError, normal carbons don't consider 14C
    print("Number of Protium atoms: " + str(conf[1][0]))
    print("Number of Deuterium atoms: " + str(conf[1][1]))
    print("Number of O16 atoms: " + str(conf[2][0]))
    print("Number of O17 atoms: " + str(conf[2][1]))
    print("Number of O18 atoms: " + str(conf[2][2]))









"""Here we store formulas for approximating numbers of isotopologues and subisotopologues."""
from math import exp, log, lgamma, pi

from .IsoSpecPy import IsoParamsFromFormula


def log_multinomial_confs_cnt(n, i):
	"""Return the natural logarithm of the number of configurations of the multinomial distribution.

	There are n independent trials of an experiment that results in one of the i outcomes.
	For each outcome, count how many times it occured.
	These counts then follow the multinomial distribution.

	Args:
		n (int): The number of trials.
		i (int): The number of possible outcomes.
	"""
	return lgamma(n + i) - \
		   lgamma(n + 1) - \
		   lgamma(i)

def test_log_multinomial_confs_cnt():
	x = log_multinomial_confs_cnt(10,3)
	assert abs(x - 4.189655) < 10**(-5)


def multinomial_confs_cnt(n, i):
	"""Return the approximate number of configurations of the multinomial distribution.

	Args:
		n (int): The number of trials.
		i (int): The number of possible outcomes.
	"""
	return exp(log_subiso_cnt_simplex(n, i))


def log_V_simplex(n, i):
	"""Get the natural logarithm of the volume of a simplex {(x_1,..,x_{i-1}): \sum_{j=1}^i = n}.

	Args:
		n (int): The number of atoms of the element.
		i (int): The number of isotopes of the element.
	"""
	return (i-1)*log(n) - lgamma(i)


def test_log_V_simplex():
	x = log_V_simplex(10,3)
	assert abs(x - 3.912023) < 10**(-5)


def V_simplex(n, i):
	"""Get the volume of a simplex {(x_1,..,x_{i-1}): \sum_{j=1}^i = n}.

	Args:
		n (int): The number of atoms of the element.
		i (int): The number of isotopes of the element.
	"""
	return exp(log_V_simplex(n,i))

def log_V_ellipsoid(n, R2, probs):
	"""Get the natural logarithm of the volume of the ellipsoid.

	The ellipsoid is defined by x' W^{-1} x <= R2,
	where W = diag(probs[0:-1]) - probs[0:-1] * probs[0:-1]'
	and R2 is the square of the radius.
	diag(probs) is a matrix with values probs[0:-1] on the diagonal,
	and probs[0:-1] * probs[0:-1]' is a projection on probs[0:-1].
	Since probabilities are nonzero and sum to one, then det W != 0.
	Also, the expression does not really depend upon the choice of one of the
	ommited probability term, e.g. the outcome would stay the same if we remover p[1].

	Args:
		n (int): The number of atoms of the element.
		R2 (float): The square of the radius of the ellipsoid.
		probs (list): List with the natural frequencies of isotopes.
	"""
	i = len(probs)
	log_det = sum(log(p) for p in probs) 
	return (log_det + (i-1)*(log(n) + log(R2) + log(pi)))/2.0 - lgamma((i+1)/2.0)


def test_log_V_ellipsoid():
	assert abs(log_V_ellipsoid(100, 10, [.2,.3,.5]) - 6.299206) < 10**(-5)


def V_ellipsoid(n, R2, probs):
	"""Get the volume of the ellipsoid."""
	return exp(log_V_ellipsoid(n, R2, probs))

def log_subisotopologue_cnt(atoms_cnt, isotope_frequencies, ellipsoid_R2):
	"""Get the natural logarithm of the approximate number of subisotopologues.

	Args:
		atoms_cnt (int): The number of atoms of the given element.
		isotope_frequencies (list): The natural frequencies of isotopes (sum to one).
		ellipsoid_R2 (float): The radius of the ellipsoid used to approximate the optimal P-set.
	"""
	isotopes_cnt = len(isotope_frequencies)
	return log_multinomial_confs_cnt(atoms_cnt, isotopes_cnt) + \
		   log_V_ellipsoid(atoms_cnt, ellipsoid_R2, isotope_frequencies) - \
		   log_V_simplex(atoms_cnt, isotopes_cnt)

def test_log_subisotopologue_cnt():
	assert abs(log_subisotopologue_cnt(100, [.2,.3,.5], 10) - 6.328959) < 10**(-5)


def subisotopologue_cnt(atoms_cnt, isotope_frequencies, ellipsoid_R2):
	"""Get the approximate number of subisotopologues."""
	return exp(log_subisotopologue_cnt(atoms_cnt, isotope_frequencies, ellipsoid_R2))


def approximate_subisotopologues(molecule, P):
	"""Approximate the number of subisotopologues.

	Args:
		molecule (str): A string with molecule, e.g. 'C100H202'.
		P (float): The joint probability threshold.
	"""
	from scipy.stats import chi2
	assert P >= 0 and P <= 1, 'That is not a probability.'
	mol = IsoParamsFromFormula(molecule)
	chi2_df = sum(len(p) for p in mol.probs) - len(mol.probs)
	R2 = chi2.ppf(q=P, df=chi2_df)
	return {e: subisotopologue_cnt(n, p, R2) for n, p, e in \
			zip(mol.atomCounts, mol.probs, mol.elems)}


if __name__ == '__main__':
	print(approximate_subisotopologues('C100H202', .999))
	print(approximate_subisotopologues("C100H100", .999))



import IsoSpecPy
from math import exp
from IsoSpecPy.Formulas import *
from IsoSpecPy.approximations import approximate_subisotopologues

test_on = horse_myoglobin
test_prob = 0.9999

print("Formula:", test_on, "Probability:", test_prob)

i = IsoSpecPy.Iso(test_on)
print ("From C++ code:", list(map(exp,i.getMarginalLogSizeEstimates(test_prob))))

symbols, _ = IsoSpecPy.ParseFormula(test_on)
dct = approximate_subisotopologues(test_on, test_prob)
print ("From Python:  ", [dct[s] for s in symbols])

v = IsoSpecPy.IsoTotalProb(formula=test_on, prob_to_cover = test_prob, get_confs = True, get_minimal_pset = True)

acc = [set() for _ in range(v.dimNumber)]

for conf in v.confs:
    for i in range(v.dimNumber):
        acc[i].add(conf[i])

print ("Real:", list(map(len, acc)))
print (len(v), "total confs.")



from __future__ import print_function
import IsoSpecPy
from IsoSpecPy.Formulas import *
import math


try:
    math.isclose
except AttributeError:
    def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
        return abs(a-b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)
    math.isclose = isclose

glu = IsoSpecPy.IsoThreshold(0.0, formula=glucose)
ca = IsoSpecPy.IsoThreshold(0.0, formula=caffeine)

print("Checking Wasserstein distance...", end=' ')
print(ca.wassersteinDistance(glu), end=' ')
assert(math.isclose(ca.wassersteinDistance(glu), 14.03495145836358))
print("OK!")

print("Checking normalization... ", end='')

ubiq = IsoSpecPy.IsoTotalProb(0.9999, ubiquitin)
print(ubiq.total_prob(), end=' ')
assert(math.isclose(ubiq.total_prob(), 0.9999, rel_tol=0.01))
ubiq = IsoSpecPy.IsoTotalProb(0.9999, ubiquitin)
ubiq.scale(0.5)
assert(math.isclose(ubiq.total_prob(), 0.9999*0.5, rel_tol=0.01))
ubiq._recalculate_everything()
assert(math.isclose(ubiq.total_prob(), 0.9999*0.5, rel_tol=0.01))
ubiq = IsoSpecPy.IsoTotalProb(0.9999, ubiquitin)
ubiq.scale(0.5)
ubiq.normalize()
assert(math.isclose(ubiq.total_prob(), 1.0))
ubiq._recalculate_everything()
assert(math.isclose(ubiq.total_prob(), 1.0))
print("OK!")


print("Checking addition...", end=' ')
wa = IsoSpecPy.IsoThreshold(0.0, formula=water)
ox = IsoSpecPy.IsoThreshold(0.0, formula=oxygen)
s = wa+ox
assert(math.isclose(s.total_prob(), 2.0))
assert(len(list(s)) == len(list(wa)) + len(list(ox)) == 15)
print("OK!")


print("Checking sorting...", end=' ')
ubiq = IsoSpecPy.IsoTotalProb(0.9999, ubiquitin)
ubiq.sort_by_mass()
assert(list(ubiq.masses) == sorted(ubiq.masses))
ubiq.sort_by_prob()
assert(list(ubiq.probs) == sorted(ubiq.probs))
print("OK!")

print("Checking binning...", end= ' ')
ubiq = IsoSpecPy.IsoTotalProb(0.999999, ubiquitin)
#ubiq.plot()
print(len(ubiq), end = ' -> ')
bu = ubiq.binned()
ubiq._recalculate_everything()
print(len(bu), end = ' ')
bu._recalculate_everything()
assert(math.isclose(ubiq.total_prob(), bu.total_prob()))
print("OK!")

print("Checking convolution...", end=' ')
o = IsoSpecPy.IsoThreshold(0.0, formula="H1")
sur = IsoSpecPy.IsoThreshold(0.0, formula=surcose)
#(glu*glu).plot()
#(sur*wa).plot()
WSD = (sur*wa).wassersteinDistance(glu*glu)
print(WSD, end=' ')
assert math.isclose(WSD, 0.0, abs_tol=1e-7)
print("OK!")


print("Checking negative formulas... ", end="")
try:
    I = Iso(formula="C-10")
    print("FAIL: exception not thrown")
except Exception as e:
    print("""exception successfully obtained, message: "{}" -> OK!""".format((str(e))))


print("Checking FASTA + negative formulas... ", end="")
try:
    I = Iso(fasta = "C", formula="C-5")
    print("FAIL: exception not thrown")
except Exception as e:
    print("""exception successfully obtained, message: "{}" -> OK!""".format((str(e))))


print("Checking FASTA + modification... ", end="")
# Selenation of methionine
I = IsoSpecPy.IsoTotalProb(0.999, formula = "C5H9N1O1Se1")
I2 = IsoSpecPy.IsoTotalProb(0.999, fasta = "M", formula = "S-1Se1")
WSD = I.wassersteinDistance(I2)
print(WSD, end="")
assert(math.isclose(I.wassersteinDistance(I2), 0.0))
print(" -> OK!")


print("Checking empiric avg mass... ", end="")
formulas = "C1 P1 H1 H100 P100 N100 O100 C100H100N100 C100O100".split()
isos = [IsoSpecPy.Iso(formula) for formula in formulas]
dists = [IsoSpecPy.IsoThreshold(0.0, formula) for formula in formulas]
diffs = [abs(i.getTheoreticalAverageMass() - d.empiric_average_mass()) for i, d in zip(isos, dists)]
print(max(diffs), end="")
assert max(diffs) < 1e-6
print(" -> OK!")


print("Checking empiric variance... ", end="")
diffs = [abs(i.variance() - d.empiric_variance()) for i, d in zip(isos, dists)]
print(max(diffs), end="")
assert max(diffs) < 1e-6
print(" -> OK!")

print("Checking empiric stddev... ", end="")
diffs = [abs(i.stddev() - d.empiric_stddev()) for i, d in zip(isos, dists)]
print(max(diffs), end="")
assert max(diffs) < 1e-6
print(" -> OK!")



import cppyy # NOTE: No import IsoSpecPy or anything like that!
#cppyy.include("unity-build.cpp")
cppyy.include("isoSpec++.h")
cppyy.load_library("libIsoSpec++.so")
i = cppyy.gbl.IsoSpec.Iso("C100H100")
iso = cppyy.gbl.IsoSpec.IsoThresholdGenerator(cppyy.gbl.std.move(i), 0.01)
while iso.advanceToNextConfiguration():
    print iso.prob(), iso.mass()



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2015-2020 Mateusz Łącki and Michał Startek.
#
#   This file is part of IsoSpec.
#
#   IsoSpec is free software: you can redistribute it and/or modify
#   it under the terms of the Simplified ("2-clause") BSD licence.
#
#   IsoSpec is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#   You should have received a copy of the Simplified BSD Licence
#   along with IsoSpec.  If not, see <https://opensource.org/licenses/BSD-2-Clause>.
# 

from .isoFFI import isoFFI
import re
import types
from . import PeriodicTbl
from .confs_passthrough import ConfsPassthrough
from collections import namedtuple, OrderedDict
import math

try:
    xrange
except NameError:
    xrange = range

regex_pattern = re.compile('([A-Z][a-z]?)(-?[0-9]*)')
ParsedFormula = namedtuple('ParsedFormula', 'atomCounts masses probs elems')




def ParseFormula(formula):
    """Parse a chemical formula.

    Args:
        formula (str): a chemical formula, e.g. "C2H6O1" or "C2H6O"
    
    Returns:
        A tuple containing element symbols and atomCounts of elements in the parsed formula.
    """
    global regex_pattern

    ret = OrderedDict()

    last = 0
    for match in re.finditer(regex_pattern, formula):
        elem, cnt = match.groups()
        if elem in ret:
            raise ValueError("""Invalid formula: {} (repeating element: "{}")""".format(formula, elem))
        ret[elem] = int(cnt) if cnt is not '' else 1
        if last!=match.start():
            raise ValueError("""Invalid formula: {}  (garbage inside: "{}")""".format(formula, formula[last:match.start()]))
        if elem not in PeriodicTbl.symbol_to_masses:
            raise ValueError("""Invalid formula: {} (unknown element symbol: "{}")""".format(formula, elem))
        last = match.end()

    if len(formula) != last:
        raise ValueError('''Invalid formula: {}  (trailing garbage: "{}")'''.format(formula, formula[last:]))

    if len(ret) == 0:
        raise ValueError("Invalid formula (empty)")

    return ret

fasta_parsing_space = isoFFI.ffi.new("int[6]")

def ParseFASTA(fasta):
    if isinstance(fasta, str):
        fasta = fasta.encode("ascii")
    isoFFI.clib.parse_fasta_c(fasta, fasta_parsing_space)
    elements = list("CHNOS")
    if fasta_parsing_space[5] > 0:
        elements.append("Se")
    od = OrderedDict()
    for i in range(len(elements)):
        od[elements[i]] = fasta_parsing_space[i]
    return od


def IsoParamsFromDict(formula, use_nominal_masses = False):
    """Produces a set of IsoSpec parameters from a chemical formula.

    Args:
        formula (dict): a parsed chemical formula, e.g. {"C": 2, "H": 6, "O": 1}
        use_nominal_masses (boolean): use masses of elements rounded to integer numbers (nominal masses)

    Returns:
        ParsedFormula: a tuple containing atomCounts, masses and marginal probabilities of elements in the parsed formula.
    """

    symbols, atomCounts = [], []
    for symbol, atomCount in formula.items():
        symbols.append(symbol)
        atomCounts.append(atomCount)

    try:
        if use_nominal_masses:
            masses = [PeriodicTbl.symbol_to_massNo[s] for s in symbols]
        else:
            masses = [PeriodicTbl.symbol_to_masses[s] for s in symbols]
        probs  = [PeriodicTbl.symbol_to_probs[s]  for s in symbols]
    except KeyError:
        raise ValueError("Invalid formula")

    return ParsedFormula(atomCounts, masses, probs, symbols)



class Iso(object):
    """Virtual class representing an isotopic distribution."""
    def __init__(self, formula="",
                 get_confs=False,
                 atomCounts=None,
                 isotopeMasses=None,
                 isotopeProbabilities=None,
                 use_nominal_masses = False,
                 fasta = ""):
        """Initialize Iso.

        Args:
            formula (str): a chemical formula, e.g. "C2H6O1" or "C2H6O".
            get_confs (boolean): should we report counts of isotopologues?
            atomCounts (list): a list of atom counts (alternative to 'formula').
            isotopeMasses (list): a list of lists of masses of elements with counts in 'atomCounts'.
            isotopeProbabilities (list): a list of lists of probabilities of elements with counts in 'atomCounts'.
            use_nominal_masses (boolean): should the masses be rounded to the closest integer values.
        """

        self.iso = None

        if len(fasta) > 0:
            molecule = ParseFASTA(fasta)
        else:
            molecule = OrderedDict()

        if len(formula) > 0:
            if isinstance(formula, dict):
                df = formula
            else:
                df = ParseFormula(formula)
            for symbol, count in df.items():
                molecule[symbol] = molecule.get(symbol, 0) + count

        for sym, cnt in molecule.items():
            if cnt < 0:
                raise Exception("Negative count of element " + sym + ": " + str(cnt))

        if len(molecule) == 0 and not all([atomCounts, isotopeMasses, isotopeProbabilities]):
            raise Exception("Either formula, fasta or ALL of: atomCounts, isotopeMasses, isotopeProbabilities must not be None")

        if len(molecule) > 0:
            self.atomCounts, self.isotopeMasses, self.isotopeProbabilities, _ = IsoParamsFromDict(molecule, use_nominal_masses = use_nominal_masses)
        else:
            self.atomCounts, self.isotopeMasses, self.isotopeProbabilities = [], [], []

        if not (atomCounts is None):
            self.atomCounts.extend(atomCounts)

        if not (isotopeMasses is None):
            self.isotopeMasses.extend(isotopeMasses)

        if not (isotopeProbabilities is None):
            self.isotopeProbabilities.extend(isotopeProbabilities)

        for sublist in self.isotopeProbabilities:
            for prob in sublist:
                if not (0.0 < prob <= 1.0):
                    raise ValueError("All isotope probabilities p must fulfill: 0.0 < p <= 1.0")

        self.isotopeNumbers = tuple(map(len, self.isotopeMasses))
        assert self.isotopeNumbers == tuple(map(len, self.isotopeProbabilities))
        assert len(self.atomCounts) == len(self.isotopeNumbers) == len(self.isotopeProbabilities)

        self.dimNumber = len(self.isotopeNumbers)

        self.get_confs = get_confs
        self.ffi = isoFFI.clib

        offsets = []

        if get_confs:
            i = 0
            for j in xrange(self.dimNumber):
                newl = []
                for k in xrange(self.isotopeNumbers[j]):
                    newl.append(i)
                    i += 1
                offsets.append(tuple(newl))
            self.offsets = tuple(offsets)

        self.iso = self.ffi.setupIso(self.dimNumber, self.isotopeNumbers,
                                     self.atomCounts,
                                     [i for s in self.isotopeMasses for i in s],
                                     [i for s in self.isotopeProbabilities for i in s])

    def __del__(self):
        try:
            if self.iso is not None:
                self.ffi.deleteIso(self.iso)
                self.iso = None
        except AttributeError:
            pass

    def getLightestPeakMass(self):
        """Get the lightest peak in the isotopic distribution."""
        return self.ffi.getLightestPeakMassIso(self.iso)

    def getHeaviestPeakMass(self):
        """Get the heaviest peak in the isotopic distribution."""
        return self.ffi.getHeaviestPeakMassIso(self.iso)

    def getMonoisotopicPeakMass(self):
        """Get the monoisotopic mass of the peak."""
        return self.ffi.getMonoisotopicPeakMassIso(self.iso)

    def getModeLProb(self):
        """Get the log probability of the most probable peak(s) in the isotopic distribution."""
        return self.ffi.getModeLProbIso(self.iso)

    def getModeMass(self):
        """Get the mass of the most probable peak.

        If there are more, return only the mass of one of them."""
        return self.ffi.getModeMassIso(self.iso)

    def getTheoreticalAverageMass(self):
        return self.ffi.getTheoreticalAverageMassIso(self.iso)

    def variance(self):
        return self.ffi.getIsoVariance(self.iso)

    def stddev(self):
        return self.ffi.getIsoStddev(self.iso)

    def getMarginalLogSizeEstimates(self, prob):
        cbuf = isoFFI.clib.getMarginalLogSizeEstimates(self.iso, prob)
        ret = list(isoFFI.ffi.cast('double[' + str(self.dimNumber) + ']', cbuf))
        isoFFI.clib.freeReleasedArray(cbuf)
        return ret

    def parse_conf(self, cptr, starting_with = 0):
        return tuple(tuple(cptr[i+starting_with] for i in o) for o in self.offsets)

    def _get_parse_conf_fun(self):
        # Can't just use the above function as lambda, as the entire class instance will be in closure
        offsets = self.offsets
        def pc(cptr, starting_with = 0):
            return tuple(tuple(cptr[i+starting_with] for i in o) for o in offsets)
        return pc


class IsoDistribution(object):
    """Isotopoic distribution with precomputed vector of masses and probabilities."""
    def np_masses(self):
        """Return computed masses as a numpy array."""
        try:
            import numpy as np
        except ImportError:
            raise Exception(e.msg + "\nThis requires numpy to be installed.")
        return np.frombuffer(isoFFI.ffi.buffer(self.masses))

    def np_probs(self):
        """Return computed probabilities as a numpy array."""
        try:
            import numpy as np
        except ImportError as e:
            raise Exception(e.msg + "\nThis requires numpy to be installed.")
        return np.frombuffer(isoFFI.ffi.buffer(self.probs))

    def __iter__(self):
        if hasattr(self, "confs") and self.confs is not None:
            for i in xrange(self.size):
                yield(self.masses[i], self.probs[i], self.confs[i])
        else:
            for i in xrange(self.size):
                yield (self.masses[i], self.probs[i])

    def __getitem__(self, idx):
        try:
            return (self.masses[idx], self.probs[idx], self.confs[idx])
        except (AttributeError, TypeError):
            return (self.masses[idx], self.probs[idx])

    def __len__(self):
        """Get the number of calculated peaks."""
        return self.size

    def _get_conf(self, idx):
        return self.parse_conf(self.raw_confs, starting_with = self.sum_isotope_numbers * idx)

    def __del__(self):
        pass

    def __init__(self, cobject = None, probs = None, masses = None, get_confs = False, iso = None):
        self.mass_sorted = False
        self.prob_sorted = False
        self._total_prob = float('nan')

        if cobject is not None:
            self.size = isoFFI.clib.confs_noFixedEnvelope(cobject)

            def wrap(typename, what, attrname, mult = 1):
                if what is not None:
                    x = isoFFI.ffi.gc(isoFFI.ffi.cast(typename + '[' + str(self.size*mult) + ']', what), isoFFI.clib.freeReleasedArray)
                    setattr(self, attrname, x)

            wrap("double", isoFFI.clib.massesFixedEnvelope(cobject), "masses")
            wrap("double", isoFFI.clib.probsFixedEnvelope(cobject), "probs")

            if get_confs:
                # Must also be a subclass of Iso...
                self.sum_isotope_numbers = sum(iso.isotopeNumbers)
                wrap("int", isoFFI.clib.confsFixedEnvelope(cobject), "raw_confs", mult = self.sum_isotope_numbers)
                self.confs = ConfsPassthrough(lambda idx: self._get_conf(idx), self.size)
                self.parse_conf = iso._get_parse_conf_fun()

        elif probs is not None and masses is not None:
            self.size = len(probs)
            type_str = "double["+str(self.size)+"]"
            self.probs  = isoFFI.ffi.new(type_str, probs)
            self.masses = isoFFI.ffi.new(type_str, masses)

    def _get_cobject(self):
        return isoFFI.clib.setupFixedEnvelope(self.masses, self.probs, len(self.masses), self.mass_sorted, self.prob_sorted, self._total_prob)

    def __add__(self, other):
        x = self._get_cobject()
        y = other._get_cobject()
        cobject = isoFFI.clib.addEnvelopes(x, y)
        isoFFI.clib.deleteFixedEnvelope(x, True)
        isoFFI.clib.deleteFixedEnvelope(y, True)
        ret = IsoDistribution(cobject = cobject)
        isoFFI.clib.deleteFixedEnvelope(cobject, False)
        return ret

    def __mul__(self, other):
        x = self._get_cobject()
        y = other._get_cobject()
        cobject = isoFFI.clib.convolveEnvelopes(x, y)
        isoFFI.clib.deleteFixedEnvelope(x, True)
        isoFFI.clib.deleteFixedEnvelope(y, True)
        ret = IsoDistribution(cobject = cobject)
        isoFFI.clib.deleteFixedEnvelope(cobject, False)
        return ret

    def total_prob(self):
        if math.isnan(self._total_prob):
            co = self._get_cobject()
            self._total_prob = isoFFI.clib.getTotalProbOfEnvelope(co)
            isoFFI.clib.deleteFixedEnvelope(co, True)
        return self._total_prob

    def normalize(self):
        co = self._get_cobject()
        isoFFI.clib.normalizeEnvelope(co)
        isoFFI.clib.deleteFixedEnvelope(co, True)
        self._total_prob = 1.0

    def scale(self, factor):
        co = self._get_cobject()
        isoFFI.clib.scaleEnvelope(co, factor)
        isoFFI.clib.deleteFixedEnvelope(co, True)
        self._total_prob *= factor

    def sort_by_prob(self):
        if not self.prob_sorted:
            co = self._get_cobject()
            isoFFI.clib.sortEnvelopeByProb(co)
            isoFFI.clib.deleteFixedEnvelope(co, True)
            self.mass_sorted = False
            self.prob_sorted = True

    def sort_by_mass(self):
        if not self.mass_sorted:
            co = self._get_cobject()
            isoFFI.clib.sortEnvelopeByMass(co)
            isoFFI.clib.deleteFixedEnvelope(co, True)
            self.mass_sorted = True
            self.prob_sorted = False

    def _recalculate_everything(self):
        self._total_prob = float('nan')
        self.mass_sorted = False
        self.prob_sorted = False

    def empiric_average_mass(self):
        co = self._get_cobject()
        ret = isoFFI.clib.empiricAverageMass(co)
        isoFFI.clib.deleteFixedEnvelope(co, True)
        return ret

    def empiric_variance(self):
        co = self._get_cobject()
        ret = isoFFI.clib.empiricVariance(co)
        isoFFI.clib.deleteFixedEnvelope(co, True)
        return ret

    def empiric_stddev(self):
        co = self._get_cobject()
        ret = isoFFI.clib.empiricStddev(co)
        isoFFI.clib.deleteFixedEnvelope(co, True)
        return ret

    def wassersteinDistance(self, other):
        x = self._get_cobject()
        y = other._get_cobject()
        ret = isoFFI.clib.wassersteinDistance(x, y)
        isoFFI.clib.deleteFixedEnvelope(x, True)
        isoFFI.clib.deleteFixedEnvelope(y, True)
        self.mass_sorted = True
        self.prob_sorted = False
        other.mass_sorted = True
        other.prob_sorted = False
        if math.isnan(ret):
            raise ValueError("Both spectra must be normalized before Wasserstein distance can be computed.")
        return ret

    def orientedWassersteinDistance(self, other):
        x = self._get_cobject()
        y = other._get_cobject()
        ret = isoFFI.clib.orientedWassersteinDistance(x, y)
        isoFFI.clib.deleteFixedEnvelope(x, True)
        isoFFI.clib.deleteFixedEnvelope(y, True)
        self.mass_sorted = True
        self.prob_sorted = False
        other.mass_sorted = True
        other.prob_sorted = False
        if math.isnan(ret):
            raise ValueError("Both spectra must be normalized before Wasserstein distance can be computed.")
        return ret

    def binned(self, width = 1.0, middle = 0.0):
        co = self._get_cobject()
        cbo = isoFFI.clib.binnedEnvelope(co, width, middle)
        isoFFI.clib.deleteFixedEnvelope(co, True)
        ret = IsoDistribution(cobject = cbo)
        isoFFI.clib.deleteFixedEnvelope(cbo, False)
        self.mass_sorted = True
        self.prob_sorted = False
        ret.mass_sorted = True
        ret.prob_sorted = False
        return ret


    @staticmethod
    def LinearCombination(envelopes, intensities):
        envelope_objs = [x._get_cobject() for x in envelopes]
        if type(intensities) != list:
            intensities = list(intensities)
        cobject = isoFFI.clib.linearCombination(envelope_objs, intensities, len(envelope_objs))
        for x in envelope_objs:
            isoFFI.clib.deleteFixedEnvelope(x, True)
        ret = IsoDistribution(cobject = cobject)
        isoFFI.clib.deleteFixedEnvelope(cobject, False)
        return ret


    def plot(self, plot_type = "bars", show = True, **matplotlib_args):
        """Plot the isotopic distribution.

        Args:
            **matplotlib_args: arguments for matplotlib plot.
        """
        try:
            from matplotlib import pyplot as plt
        except ImportError as e:
            raise ImportError(str(e) + "\nPlotting spectra requires matplotlib to be installed.")
        if plot_type == "bars":
            plt.vlines(list(self.masses), [0], list(self.probs), linewidth=1, **matplotlib_args)
        elif plot_type == "profile":
            self.sort_by_mass()
            plt.plot(list(self.masses), list(self.probs), **matplotlib_args)
        plt.xlabel("Mass (Da)")
        plt.ylabel("Intensity (relative)")
        if show:
            plt.show()




def IsoThreshold(threshold,
                 formula="",
                 absolute=False,
                 get_confs=False,
                 **kwargs):
    """Initialize the IsoDistribution isotopic distribution by threshold.

    Args:
        threshold (float): value of the absolute or relative threshold.
        formula (str): a chemical formula, e.g. "C2H6O1" or "C2H6O".
        absolute (boolean): should we report peaks with probabilities above an absolute probability threshold, or above a relative threshold amounting to a given proportion of the most probable peak?
        get_confs (boolean): should we report counts of isotopologues?
        **kwds: named arguments to IsoSpectrum.
    """
    iso = Iso(formula = formula, get_confs = get_confs, **kwargs)
    tabulator = isoFFI.clib.setupThresholdFixedEnvelope(iso.iso, threshold, absolute, get_confs)
    ido = IsoDistribution(cobject = tabulator, get_confs = get_confs, iso = iso)
    isoFFI.clib.deleteFixedEnvelope(tabulator, False)
    return ido


def IsoTotalProb(prob_to_cover,
                 formula="",
                 get_minimal_pset=True,
                 get_confs=False,
                 **kwargs):
        """Initialize the IsoDistribution isotopic distribution by total probability.

        Args:
            prob_to_cover (float): minimal total probability of the reported peaks.
            formula (str): a chemical formula, e.g. "C2H6O1" or "C2H6O".
            get_minimal_pset (boolean): should we trim the last calculated layer of isotopologues so that the reported result is as small as possible?
            get_confs (boolean): should we report the counts of isotopologues?
            **kwargs: named arguments to the superclass.
        """
        iso = Iso(formula=formula, get_confs=get_confs, **kwargs)
        tabulator = isoFFI.clib.setupTotalProbFixedEnvelope(iso.iso, prob_to_cover, get_minimal_pset, get_confs)
        ido = IsoDistribution(cobject = tabulator, get_confs = get_confs, iso = iso)
        isoFFI.clib.deleteFixedEnvelope(tabulator, False)
        return ido


def IsoStochastic(no_molecules,
                 formula="",
                 precision=0.9999,
                 beta_bias=5.0,
                 get_confs=False,
                 **kwargs):
        """Initialize the IsoDistribution isotopic distribution by total probability.

        Args:
            no_molecules (uint): ionic current in instrument
            formula (str): a chemical formula, e.g. "C2H6O1" or "C2H6O".
            precision (float): passed to IsoTotalProbGenerator. Between 0.0 and 1.0.
            beta_bias (float, nonnegative): fiddling with this parameter does not change the result, but might make computations slightly faster (or likely, much, much slower is you screw it up...)
            get_confs (boolean): should we report the counts of isotopologues?
            **kwargs: named arguments to the superclass.
        """
        iso = Iso(formula=formula, get_confs=get_confs, **kwargs)
        tabulator = isoFFI.clib.setupStochasticFixedEnvelope(iso.iso, no_molecules, precision, beta_bias, get_confs)
        ido = IsoDistribution(cobject = tabulator, get_confs = get_confs, iso = iso)
        isoFFI.clib.deleteFixedEnvelope(tabulator, False)
        return ido


def IsoBinned(bin_width,
                 formula="",
                 target_total_prob=0.9999,
                 bin_middle=0.0,
                 **kwargs):
        """Initialize the IsoDistribution isotopic distribution by total probability.

        Args:
            no_molecules (uint): ionic current in instrument
            formula (str): a chemical formula, e.g. "C2H6O1" or "C2H6O".
            precision (float): passed to IsoTotalProbGenerator. Between 0.0 and 1.0.
            beta_bias (float, nonnegative): fiddling with this parameter does not change the result, but might make computations slightly faster (or likely, much, much slower is you screw it up...)
            get_confs (boolean): should we report the counts of isotopologues?
            **kwargs: named arguments to the superclass.
        """
        iso = Iso(formula=formula, get_confs=False, **kwargs)
        tabulator = isoFFI.clib.setupBinnedFixedEnvelope(iso.iso, target_total_prob, bin_width, bin_middle)
        ido = IsoDistribution(cobject = tabulator, get_confs = False, iso = iso)
        isoFFI.clib.deleteFixedEnvelope(tabulator, False)
        return ido


class IsoGenerator(Iso):
    """Virtual class alowing memory-efficient iteration over the isotopic distribution.

    This iterator will stop only after enumerating all isotopologues.
    """
    def __init__(self, formula="", get_confs=False, **kwargs):
        """Initialize the IsoGenerator.

        Args:
            formula (str): a chemical formula, e.g. "C2H6O1" or "C2H6O".
            get_confs (boolean): should we report the counts of isotopologues?
            **kwargs: named arguments to the superclass.
        """
        self.cgen = None
        super(IsoGenerator, self).__init__(formula=formula, get_confs=get_confs, **kwargs)
        self.conf_space = isoFFI.ffi.new("int[" + str(sum(self.isotopeNumbers)) + "]")
        self.firstuse = True

    def __iter__(self):
        if not self.firstuse:
            raise NotImplementedError("Multiple iterations through the same IsoGenerator object are not supported. Either create a new (identical) generator for a second loop-through, or use one of the non-generator classes, which do support being re-used.")
        self.firstuse = False
        cgen = self.cgen
        if self.get_confs:
            while self.advancer(cgen):
                self.conf_getter(cgen, self.conf_space)
                yield (self.mass_getter(cgen), self.xprob_getter(cgen), self.parse_conf(self.conf_space))
        else:
            while self.advancer(cgen):
                yield (self.mass_getter(cgen), self.xprob_getter(cgen))

    def __del__(self):
        super(IsoGenerator, self).__del__()
        

class IsoThresholdGenerator(IsoGenerator):
    """Class alowing memory-efficient iteration over the isotopic distribution up till some probability threshold.

    This iterator will stop only after reaching a probability threshold.
    """
    def __init__(self, threshold, formula="", absolute=False, get_confs=False, reorder_marginals = True, **kwargs):
        """Initialize IsoThresholdGenerator.

        Args:
            formula (str): a chemical formula, e.g. "C2H6O1" or "C2H6O".
            absolute (boolean): should we report peaks with probabilities above an absolute probability threshold, or above a relative threshold amounting to a given proportion of the most probable peak?
            get_confs (boolean): should we report the counts of isotopologues?
            **kwargs: named arguments to the superclass.
        """
        super(IsoThresholdGenerator, self).__init__(formula=formula, get_confs=get_confs, **kwargs)
        self.threshold = threshold
        self.absolute = absolute
        self.cgen = self.ffi.setupIsoThresholdGenerator(self.iso,
                                                        threshold,
                                                        absolute,
                                                        1000,
                                                        1000,
                                                        reorder_marginals)
        self.advancer = self.ffi.advanceToNextConfigurationIsoThresholdGenerator
        self.xprob_getter = self.ffi.probIsoThresholdGenerator
        self.mass_getter = self.ffi.massIsoThresholdGenerator
        self.conf_getter = self.ffi.get_conf_signatureIsoThresholdGenerator

    def __del__(self):
        """Destructor."""
        try:
            if self.cgen is not None:
                self.ffi.deleteIsoThresholdGenerator(self.cgen)
                self.cgen = None
        except AttributeError:
            pass
        super(IsoThresholdGenerator, self).__del__()


class IsoLayeredGenerator(IsoGenerator):
    """Class alowing memory-efficient iteration over the isotopic distribution up till some joint probability of the reported peaks."""
    def __init__(self, formula="", get_confs=False, reorder_marginals = True, t_prob_hint = 0.99, **kwargs):
        """Initialize IsoThresholdGenerator.

        Args:
            formula (str): a chemical formula, e.g. "C2H6O1" or "C2H6O".
            absolute (boolean): should we report peaks with probabilities above an absolute probability threshold, or above a relative threshold amounting to a given proportion of the most probable peak?
            get_confs (boolean): should we report the counts of isotopologues?
            **kwargs: named arguments to the superclass.
        """
        super(IsoLayeredGenerator, self).__init__(formula=formula, get_confs=get_confs, **kwargs)
        self.cgen = self.ffi.setupIsoLayeredGenerator(self.iso, 1000, 1000, reorder_marginals, t_prob_hint)
        self.advancer = self.ffi.advanceToNextConfigurationIsoLayeredGenerator
        self.xprob_getter = self.ffi.probIsoLayeredGenerator
        self.mass_getter = self.ffi.massIsoLayeredGenerator
        self.conf_getter = self.ffi.get_conf_signatureIsoLayeredGenerator

    def __del__(self):
        try:
            if self.cgen is not None:
                self.ffi.deleteIsoLayeredGenerator(self.cgen)
                self.cgen = None
        except AttributeError:
            pass
        super(IsoLayeredGenerator, self).__del__()

class IsoOrderedGenerator(IsoGenerator):
    """Class representing an isotopic distribution with peaks ordered by descending probability.

    This generator return probilities ordered with descending probability.
    It it not optimal to do so, but it might be useful.

    WARNING! This algorithm work in O(N*log(N)) vs O(N) of the threshold and layered algorithms.
    Also, the order of descending probability will most likely not reflect the order of ascending masses.
    """
    def __init__(self, formula="", get_confs=False, **kwargs):
        """Initialize IsoOrderedGenerator.

        Args:
            formula (str): a chemical formula, e.g. "C2H6O1" or "C2H6O".
            get_confs (boolean): should we report the counts of isotopologues?
            **kwargs: named arguments to the superclass.
        """
        super(IsoOrderedGenerator, self).__init__(formula=formula, get_confs=get_confs, **kwargs)
        self.cgen = self.ffi.setupIsoOrderedGenerator(self.iso, 1000, 1000)
        self.advancer = self.ffi.advanceToNextConfigurationIsoOrderedGenerator
        self.xprob_getter = self.ffi.probIsoOrderedGenerator
        self.mass_getter = self.ffi.massIsoOrderedGenerator
        self.conf_getter = self.ffi.get_conf_signatureIsoOrderedGenerator

    def __del__(self):
        try:
            if self.cgen is not None:
                self.ffi.deleteIsoLayeredGenerator(self.cgen)
                self.cgen = None
        except AttributeError:
            pass
        super(IsoOrderedGenerator, self).__del__()


class IsoStochasticGenerator(IsoGenerator):
    def __init__(self, no_molecules, formula="", precision=0.9999, beta_bias = 1.0, get_confs=False, **kwargs):
        super(IsoStochasticGenerator, self).__init__(formula=formula, get_confs=get_confs, **kwargs)
        self.threshold = precision
        self.no_molecules = no_molecules
        self.cgen = self.ffi.setupIsoStochasticGenerator(self.iso,
                                                        no_molecules,
                                                        precision,
                                                        beta_bias)
        self.advancer = self.ffi.advanceToNextConfigurationIsoStochasticGenerator
        self.xprob_getter = self.ffi.probIsoStochasticGenerator
        self.mass_getter = self.ffi.massIsoStochasticGenerator
        self.conf_getter = self.ffi.get_conf_signatureIsoStochasticGenerator

    def __del__(self):
        """Destructor."""
        try:
            if self.cgen is not None:
                self.ffi.deleteIsoStochasticGenerator(self.cgen)
                self.cgen = None
        except AttributeError:
            pass
        super(IsoStochasticGenerator, self).__del__()




import sys
import IsoSpecPy
from math import exp, log

def sprint(s):
    sys.stdout.write(str(s))
    sys.stdout.flush()


silentish_run = False

if len(sys.argv) < 2:
    try:
        import OldIsoSpecPy
    except ImportError:
        print("This test compares the results of installed IsoSpec with IsoSpec version 1.0.7, installed as OldIsoSpecPy")
        print("You must install it, using:")
        print("pip install OldIsoSpecPy --index-url https://test.pypi.org/simple/")
        print("Or run only the new algo: python " + sys.argv[0] + " True")
        sys.exit(1)
else:
    print("Running only new algo...")
    silentish_run = True

# Correctness tests comparing IsoSpecPy 1.0.7 and HEAD. The trouble here is that due to slightly different way things are calculated
# we get different rounding errors, and therefore slightly different results - and that's OK. The function kinda_like reflects that
# - but is still WAAAY too strict. Often we can justifiably get different configurations, or different counts of configurations...

molecules = "H2O1 C100 P1 P100 C1 H10C10O10N10S5 Se1 Se10 Sn1 Sn4 Sn4C1 C2H6O1 C1000 C520H817N139O147S8 C1H1O2N2Se1Sn1P1 P1C1Sn1 Se5 Sn5 Se50 Sn15 Se2Sn2C2O2N2S2B2He2U2Na2Cl2".split()

parameters = list(map(float, "0.0 0.1 0.5 0.01 0.9 0.99 0.01 0.0001 0.999 0.362 0.852348".split()))

def kinda_like(o1, o2):
    if type(o1) in (list, tuple) and type(o2) in (list, tuple) :
        assert len(o1) == len(o2)
        assert all(kinda_like(oo1, oo2) for oo1, oo2 in zip(o1, o2))
    if type(o1) == type(o2) == float:
        assert o1*o2 >= 0.0 # same sign check
        assert abs(o1*0.99)-0.000001 <= abs(o2) <= abs(o1*1.01)+0.000001
    return True

def sort_confs(confs):
    if len(confs[0]) == 0:
        return confs
    assert len(set(tuple(x) for x in confs[2])) == len(confs[2])
    l = list(zip(*confs))
    l.sort(key = lambda x: -x[1])

    return ([x[0] for x in l], [x[1] for x in l], [x[2] for x in l])

def confs_from_ordered_generator(formula, target_prob):
    ret = ([], [], [])
    prob = 0.0
    for conf in IsoSpecPy.IsoOrderedGenerator(formula=formula, get_confs=True):
        conf = (conf[0], log(conf[1]), conf[2])
        if prob >= target_prob and target_prob < 1.0:
            return ret
        ret[0].append(conf[0])
        prob += exp(conf[1])
        ret[1].append(conf[1])
        ret[2].append([item for sublist in conf[2] for item in sublist])
    return ret

def confs_from_layered_generator(formula, target_prob):
    ret = ([], [], [])
    for conf in IsoSpecPy.IsoLayered(formula=formula, prob_to_cover = target_prob, get_confs=True, get_minimal_pset=True):
        conf = (conf[0], log(conf[1]), conf[2])
        ret[0].append(conf[0])
        ret[1].append(conf[1])
        ret[2].append([item for sublist in conf[2] for item in sublist])

    return sort_confs(ret)

def confs_from_threshold_generator(formula, target_prob):
    ret = ([], [], [])
    for conf in IsoSpecPy.IsoThresholdGenerator(formula=formula, threshold = target_prob, absolute = True, get_confs=True):
        conf = (conf[0], log(conf[1]), conf[2])
        ret[0].append(conf[0])
        ret[1].append(conf[1])
        ret[2].append([item for sublist in conf[2] for item in sublist])

    return sort_confs(ret)

def confs_from_threshold(formula, target_prob):
    ret = ([], [], [])
    for conf in IsoSpecPy.IsoThreshold(formula=formula, threshold = target_prob, absolute = True, get_confs=True):
        conf = (conf[0], log(conf[1]), conf[2])
        ret[0].append(conf[0])
        ret[1].append(conf[1])
        ret[2].append([item for sublist in conf[2] for item in sublist])

    return sort_confs(ret)



is_ok = False
try:
    i = IsoSpecPy.IsoThreshold(0.1, atomCounts = [100], isotopeMasses = [[1.0, 2.0, 3.0]], isotopeProbabilities = [[0.0, 0.6, 0.4]])
    for x in i:
        print(x)
except ValueError:
    is_ok = True
assert is_ok

total_confs = 0

for molecule in molecules:
    for parameter in parameters:
        if not silentish_run:
            sprint("{} {}... ".format(molecule, parameter))
            old_ordered = OldIsoSpecPy.IsoSpecPy.IsoSpec.IsoFromFormula(molecule, parameter, method="ordered").getConfs()
            sprint(len(old_ordered[0]))
        new_ordered = confs_from_ordered_generator(molecule, parameter)
        total_confs += len(new_ordered[0])
        if not silentish_run: assert kinda_like(new_ordered, old_ordered)
        new_layered = confs_from_layered_generator(molecule, parameter)
        total_confs += len(new_layered[0])
        if not silentish_run: assert kinda_like(new_layered, new_ordered)
        if len(new_ordered[1]) > 0:
            new_threshold = exp(new_ordered[1][-1]-0.00000001)
        else:
            new_threshold = 1.1
        new_threshold_res = confs_from_threshold_generator(molecule, new_threshold)
        total_confs += len(new_threshold_res[0])
        if not silentish_run:
            assert kinda_like(new_ordered, new_threshold_res)
            assert kinda_like(new_threshold_res, confs_from_threshold(molecule, new_threshold))

        if parameter > 0:
            if not silentish_run: sprint(" thresholded: ")
            new_thr_r = confs_from_threshold_generator(molecule, parameter)
            if not silentish_run:
                sprint(len(new_thr_r[0]))
                old_thr_r = sort_confs(OldIsoSpecPy.IsoSpecPy.IsoSpec.IsoFromFormula(molecule, parameter, method="threshold_absolute").getConfs())
                assert kinda_like(new_thr_r, old_thr_r)

        if not silentish_run:
            print("... OK!")

sprint("Total confs: " + str(total_confs) + "\n")



from IsoSpecPy import IsoThreshold, Iso, ParseFormula, Advanced






def check_estimates(formula, est_fun, thr):
    it = IsoThreshold(thr, formula, absolute = False)
    min_m = min(it.masses)
    max_m = max(it.masses)

    min_e_m, max_e_m = est_fun(formula, thr)

#    assert min_e_m < min_m
#    assert max_e_m > max_e

    return max_e_m, max_m, min_m, min_e_m, max_e_m > max_m >= min_m > min_e_m, (max_e_m - min_e_m) / (max_m - min_m + 0.0001), (max_e_m - min_e_m), (max_m - min_m)



def est_fun_trivial(formula, _):
    i = Iso(formula)
    return i.getLightestPeakMass(), i.getHeaviestPeakMass()


def est_fun_margs(formula, threshold):
    PC = ParseFormula(formula)
    it = IsoThreshold(threshold, formula, get_confs = True)
    marginals = [Iso(elem, cnt) for elem, cnt in zip(PC[0], PC[1])]
    mconfs = [set() for _ in marginals]
    for conf in it.confs:
        for marg_conf, marg_set in zip(conf, mconfs):
            marg_set.add(marg_conf)

    mins  = [min(marginal.conf_mass((conf,)) for conf in cset) for cset, marginal in zip(mconfs, marginals)]
    maxes = [max(marginal.conf_mass((conf,)) for conf in cset) for cset, marginal in zip(mconfs, marginals)]

    return sum(mins) - 1.0, sum(maxes) + 1.0


from IsoSpecPy.Formulas import *

for formula in "bovine_insulin horse_myoglobin surcose water ubiquitin caffeine averagine(100.0) averagine(1000.0) averagine(100000.0) Hg10Sn5 H1000N1000".split():
    try:
        fformula = eval(formula)
    except NameError:
        fformula = formula
    for thr in [0.1, 0.01, 0.001, 0.00001]:
        print(formula, thr, check_estimates(fformula, est_fun_margs, thr))




'''This file contains a set of functions useful for the analysis of properties of
isotopic envelopes. These are unlikely to be of any interest to end-users, only
for developers interested in testing some properties of the isotopic envelopes.
These are implemented mostly in Python, and so, not as fast as the rest of the
package.'''

def _int_neighbours(subisotopologue_conf):
    conf = list(subisotopologue_conf)
    for i in range(len(conf)):
        for j in range(len(conf)):
            if i != j and conf[i] > 0:
                conf[i] -= 1
                conf[j] += 1
                yield tuple(conf)
                conf[i] += 1
                conf[j] -= 1


def neighbours(conf):
    for i in range(len(conf)):
        for si_neigh in _int_neighbours(conf[i]):
            yield tuple(conf[j] if j != i else si_neigh for j in range(len(conf)))




def conf_mass(iso, conf):
    return sum(sum(mass * cnt for mass, cnt in zip(masses, cnts)) for masses, cnts in zip(iso.isotopeMasses, conf))


from IsoSpecPy import Iso
Iso.conf_mass = conf_mass



from IsoSpecPy.IsoSpecPy import IsoTotalProb, IsoThreshold
from IsoSpecPy.PeriodicTbl import symbol_to_masses


# Suppose we want to simulate the spectrum of
# C100H200Ag46
# except that the isotope counts of silver in each molecule
# are fixed: each molecule contains exactly 44 atoms of 107Ag,
# and 2 atoms of 109Ag. The carbon and hydrogen follow their
# natural distributions.

# Here's how to do that:

# We start by defining two artificial elements, each representing
# one isotope of silver. The element is going to be monoisotopic,
# with proper mass retrieved from the periodic table.
# The new elements will be passed to Iso* methods using the
# atomCounts, isotopeMasses, isotopeProbabilities parameters,
# and the rest of the formula - using standard "formula"
# argument.

# defining extra elements
# 44 107Ag atoms and 2 109Ag
atomCounts = [44, 2]

# Retrieving the isotope masses from IsoSpec's builtin periodic table
# (one could input them manually too)
M107Ag, M109Ag = symbol_to_masses['Ag']

# The masses are passed as list of masses for each new element:
# we have two elements, each has one isotope, hence the following
# list structure
isotopeMasses = [[M107Ag], [M109Ag]]

# The single isotope of each new element occurs with 1.0 probability
isotopeProbabilities = [[1.0], [1.0]]


# peaks above .001 threshold
print("Here is the list of peaks above 0.001 intensity threshold for C100H200[107Ag]44[109Ag]2")
MZI_thr = IsoThreshold(.001,
                       'C100H200', # the remaining part of the formula
                       get_confs=True,
                       atomCounts=atomCounts,
                       isotopeMasses=isotopeMasses,
                       isotopeProbabilities=isotopeProbabilities)

print("Masses list:", list(MZI_thr.masses))
print("Probabilities list:", list(MZI_thr.probs))
print("Configurations list:", list(MZI_thr.confs))

print()
print()

# Alternatively: peaks that make 99.9% of the isotopic distribution
print("Here is the list of peaks that comprise at least 99.9% of total intensity for C100H200[107Ag]44[109Ag]2")
MZI_opt = IsoTotalProb(.999,
                       'C100H200', # the remaining part of the formula
                       get_confs=True,
                       atomCounts=atomCounts,
                       isotopeMasses=isotopeMasses,
                       isotopeProbabilities=isotopeProbabilities)

print("Masses list:", list(MZI_opt.masses))
print("Probabilities list:", list(MZI_opt.probs))
print("Configurations list:", list(MZI_opt.confs))



import math

from .IsoSpecPy import IsoDistribution




def simple_inverse(fun, prec = 0.01):
    
    def inverse(x):
        start = -1.0
        while fun(start) > x:
            start *= 2.0
        end = 1.0
        while fun(end) < x:
            end *= 2.0

        while (end - start) > prec:
            print(start, end)
            mid = (end + start) * 0.5
            if fun(mid) < x:
                start = mid
            else:
                end = mid

        return (end + start) * 0.5

    return inverse
            


class Distribution(IsoDistribution):
    def __init__(self, cdf, bin_width = 0.01, precision = 0.99, inverse_cdf = None):
        if inverse_cdf is None:
            inverse_cdf = simple_inverse(cdf, prec = bin_width*0.125)

        prec_missing = 0.5 * (1.0 - precision)

        start = inverse_cdf(prec_missing)
        end = inverse_cdf(1.0 - prec_missing)

        half_bin_width = 0.5*bin_width

        bw_inverse = 1.0 / bin_width

        def bw_round(x):
            return math.floor(x*bw_inverse + 0.5) / bw_inverse

        start = bw_round(start)
        end = bw_round(end)

        last_bin_start = start - half_bin_width
        last_cdf = cdf(last_bin_start)
        
        probs = [0.0]
        masses = [last_bin_start - half_bin_width]

        while last_bin_start < end:
            next_bin_start = last_bin_start + bin_width
            next_cdf = cdf(next_bin_start)

            masses.append(last_bin_start + half_bin_width)
            probs.append(next_cdf - last_cdf)

            last_cdf = next_cdf
            last_bin_start = next_bin_start

        sprobs = 1.0/sum(probs)
        probs = [p*sprobs for p in probs]

        probs.append(0.0)
        masses.append(last_bin_start + half_bin_width)

        super(Distribution, self).__init__(masses = masses, probs = probs)

        self.mass_sorted = True



class Gaussian(Distribution):
    def __init__(self, stdev, bin_width = 0.01, precision = 0.99):
        varfactor = 1.0/(stdev * 1.4142135623730951)
        cdf = lambda x: 0.5*(1.0+math.erf(x*varfactor))
        inv_cdf = None
        try:
            # If scipy is available, use its ppf
            from scipy.stats import norm
            inv_cdf = lambda x: norm.ppf(x, loc = 0.0, scale = stdev)
        except ImportError:
            pass # oh well, no scipy, we'll just use binsearch-based inversion

        super(Gaussian, self).__init__(cdf = cdf, bin_width = bin_width, precision = precision, inverse_cdf = inv_cdf)





import cffi
import os
import platform
import sys
import glob


class IsoFFI:
    def __init__(self):
        self.ffi = cffi.FFI()
        self.ffi.cdef('''
        void * setupIso(int dimNumber,
                const int* isotopeNumbers,
                const int* atomCounts,
                const double* isotopeMasses,
                const double* isotopeProbabilities);

        void * isoFromFasta(const char* fasta, bool use_nominal_masses, bool add_water);

        double getLightestPeakMassIso(void* iso);
        double getHeaviestPeakMassIso(void* iso);
        double getMonoisotopicPeakMassIso(void* iso);
        double getModeLProbIso(void* iso);
        double getModeMassIso(void* iso);
        double getTheoreticalAverageMassIso(void* iso);
        double getIsoVariance(void* iso);
        double getIsoStddev(void* iso);
        double* getMarginalLogSizeEstimates(void* iso, double target_total_prob);

        void deleteIso(void* iso);

        void* setupIsoThresholdGenerator(void* iso,
                                         double threshold,
                                         bool _absolute,
                                         int _tabSize,
                                         int _hashSize,
                                         bool reorder_marginals);
        double massIsoThresholdGenerator(void* generator); double lprobIsoThresholdGenerator(void* generator); double probIsoThresholdGenerator(void* generator); void methodIsoThresholdGenerator(void* generator); bool advanceToNextConfigurationIsoThresholdGenerator(void* generator); void deleteIsoThresholdGenerator(void* generator); void get_conf_signatureIsoThresholdGenerator(void* generator, int* space);



        void* setupIsoLayeredGenerator(void* iso,
                                       int _tabSize,
                                       int _hashSize,
                                       bool reorder_marginals,
                                       double t_prob_hint);
        double massIsoLayeredGenerator(void* generator); double lprobIsoLayeredGenerator(void* generator); double probIsoLayeredGenerator(void* generator); void methodIsoLayeredGenerator(void* generator); bool advanceToNextConfigurationIsoLayeredGenerator(void* generator); void deleteIsoLayeredGenerator(void* generator); void get_conf_signatureIsoLayeredGenerator(void* generator, int* space);


        void* setupIsoOrderedGenerator(void* iso,
                                       int _tabSize,
                                       int _hashSize);
        double massIsoOrderedGenerator(void* generator); double lprobIsoOrderedGenerator(void* generator); double probIsoOrderedGenerator(void* generator); void methodIsoOrderedGenerator(void* generator); bool advanceToNextConfigurationIsoOrderedGenerator(void* generator); void deleteIsoOrderedGenerator(void* generator); void get_conf_signatureIsoOrderedGenerator(void* generator, int* space);

        void* setupIsoStochasticGenerator(void* iso,
                                   size_t no_molecules,
                                   double precision,
                                   double beta_bias);
        double massIsoStochasticGenerator(void* generator); double lprobIsoStochasticGenerator(void* generator); double probIsoStochasticGenerator(void* generator); void methodIsoStochasticGenerator(void* generator); bool advanceToNextConfigurationIsoStochasticGenerator(void* generator); void deleteIsoStochasticGenerator(void* generator); void get_conf_signatureIsoStochasticGenerator(void* generator, int* space);

        void* setupThresholdFixedEnvelope(void* iso,
                                    double threshold,
                                    bool absolute,
                                    bool get_confs);

        void* setupTotalProbFixedEnvelope(void* iso,
                                      double taget_coverage,
                                      bool optimize,
                                      bool get_confs);

        void* setupStochasticFixedEnvelope(void* iso,
                                     size_t no_molecules,
                                     double precision,
                                     double beta_bias,
                                     bool get_confs);

        void* setupBinnedFixedEnvelope(void* iso,
                            double target_total_prob,
                            double bin_width,
                            double bin_middle);

        void* setupFixedEnvelope(double* masses, double* probs, size_t size, bool mass_sorted, bool prob_sorted, double total_prob);
        void deleteFixedEnvelope(void* tabulator, bool releaseEverything);

        const double* massesFixedEnvelope(void* tabulator);
        const double* probsFixedEnvelope(void* tabulator);
        const int*    confsFixedEnvelope(void* tabulator);
        size_t confs_noFixedEnvelope(void* tabulator);

        double empiricAverageMass(void* tabulator);
        double empiricVariance(void* tabulator);
        double empiricStddev(void* tabulator);

        double wassersteinDistance(void* tabulator1, void* tabulator2);
        double orientedWassersteinDistance(void* tabulator1, void* tabulator2);

        void* addEnvelopes(void* tabulator1, void* tabulator2);
        void* convolveEnvelopes(void* tabulator1, void* tabulator2);

        double getTotalProbOfEnvelope(void* envelope);
        void scaleEnvelope(void* envelope, double factor);
        void normalizeEnvelope(void* envelope);
        void* binnedEnvelope(void* envelope, double width, double middle);
        void* linearCombination(void* const * const envelopes, const double* intensities, size_t count);

        void sortEnvelopeByMass(void* envelope);
        void sortEnvelopeByProb(void* envelope);

        void freeReleasedArray(void* array);
        void parse_fasta_c(const char* fasta, int atomCounts[6]);

        #define NUMBER_OF_ISOTOPIC_ENTRIES 292
        extern const size_t isospec_number_of_isotopic_entries;
        extern const int elem_table_atomicNo[NUMBER_OF_ISOTOPIC_ENTRIES];
        extern const double elem_table_probability[NUMBER_OF_ISOTOPIC_ENTRIES];
        extern const double elem_table_mass[NUMBER_OF_ISOTOPIC_ENTRIES];
        extern const double elem_table_massNo[NUMBER_OF_ISOTOPIC_ENTRIES];
        extern const int elem_table_extraNeutrons[NUMBER_OF_ISOTOPIC_ENTRIES];
        extern const char* elem_table_element[NUMBER_OF_ISOTOPIC_ENTRIES];
        extern const char* elem_table_symbol[NUMBER_OF_ISOTOPIC_ENTRIES];
        extern const bool elem_table_Radioactive[NUMBER_OF_ISOTOPIC_ENTRIES];
                        ''');

        mod_dir = os.path.dirname(os.path.abspath(__file__))

        if os.path.exists(os.path.join(mod_dir, '..', 'setup.py')):
            raise ImportError('''Attempted to load IsoSpecPy module from its build directory. This usually
won't work and is generally a Bad Idea. Please cd somewhere else, or, if you're really
sure you want to do that, edit the source and disable this check.''')

        libnames  = ['IsoSpecCppPy*', 'IsoSpec++*']
        libprefix = ['', 'lib', 'Lib']
        extension = ['.so', '.dylib', '.dll']
        try:
            if platform.system() == 'Linux':
                extension = ['.so']
            elif platform.system() == 'Windows':
                extension = ['.dll']
        except:
            pass

        prebuilt =  ['', 'prebuilt-']

        def cprod(ll1, ll2):
            res = []
            for l1 in ll1:
                for l2 in ll2:
                    res.append(l1+l2)
            return res

        paths_to_check = cprod(prebuilt, cprod(libprefix, cprod(libnames, extension)))

        dpc = []

        for dirpath in [mod_dir, mod_dir + '/..']:
            dpc.extend([os.path.join(dirpath, p) for p in paths_to_check])

        paths_to_check = dpc

        paths_to_check = sum(map(glob.glob, paths_to_check), [])

        try:
            import importlib
            paths_to_check.insert(0, importlib.util.find_spec("IsoSpecCppPy").origin)
        except (ImportError, AttributeError):
            pass

        errors = []

        self.clib = None
        for libpath in set(paths_to_check):
            try:
                self.clib = self.ffi.dlopen(libpath)
                break
            except (IndexError, OSError) as e:
                errmsg = "Load libIsoSpec++.so, tried: " + libpath + '\n' + "Got error: " + str(type(e)) + ": " + str(e)
                errors.append(errmsg)

        if self.clib == None:
            raise ImportError("Cannot find or load the C++ part of the library\n" + '\n'.join(errors))


isoFFI = IsoFFI()  # This is done while including the module



from .isoFFI import isoFFI
from collections import defaultdict

try:
    xrange
except NameError:
    xrange = range

number_of_isotopic_entries = isoFFI.clib.NUMBER_OF_ISOTOPIC_ENTRIES

symbol_to_masses = defaultdict(tuple)
symbol_to_massNo = defaultdict(tuple)
symbol_to_probs  = defaultdict(tuple)
symbol_to_atomic_number = {}

for i in xrange(number_of_isotopic_entries):
    symbol = isoFFI.ffi.string(isoFFI.clib.elem_table_symbol[i]).decode("ascii")
    symbol_to_masses[symbol] += (isoFFI.clib.elem_table_mass[i],)
    symbol_to_massNo[symbol] += (isoFFI.clib.elem_table_massNo[i],)
    symbol_to_probs[symbol] += (isoFFI.clib.elem_table_probability[i],)
    symbol_to_atomic_number[symbol] = isoFFI.clib.elem_table_atomicNo[i]

symbol_to_masses = dict(symbol_to_masses)
symbol_to_probs = dict(symbol_to_probs)

# Several derivative convenience dicts...
symbol_to_massprob = dict((key, [zip(symbol_to_masses[key], symbol_to_probs[key])]) for key in symbol_to_probs.keys())

def crossprod(l1, l2):
    return sum(x1*x2 for x1, x2 in zip(l1, l2))

symbol_to_avg_mass = dict((key, crossprod(symbol_to_masses[key], symbol_to_probs[key])) for key in symbol_to_probs.keys())

def maxprod(l1, l2):
    return max(zip(l1, l2), key = lambda x: x[1])[0]

symbol_to_monoisotopic_mass = dict((key, maxprod(symbol_to_masses[key], symbol_to_probs[key])) for key in symbol_to_probs.keys())




import IsoSpecPy


from random import uniform, choice
from numpy.random import binomial
from math import inf, nan


def _beta_1_b(b):
    '''Returns a random variate from beta(1, b) distribution
    using the inverse CDF method'''
    return 1.0-uniform(0.0, 1.0)**(1.0/b)

def _safe_binom(n, p):
    '''Draw a sample from binomial distribution. Doesn't crash
    if p > 1.0 due to numerical inaccuracies.'''
    if p >= 1.0:
        return n
    return binomial(n, p)





#def _sample_with_replacement_online_impl(population, probabilities, sample_size):
def sample_isospec(formula, count, precision):
    population = IsoSpecPy.IsoLayeredGenerator(formula, t_prob_hint = precision, reorder_marginals = False)
    #population = IsoSpecPy.IsoThresholdGenerator(formula = formula, threshold = -1.0)

    #for x in population:
    #    yield x
    '''Performs sampling with replacement from population argument, with
    associated probabilities from second argument. The probabilities must 
    sum to 1. Yields a stream of tuples: (population_member, times_chosen).
    Accepts generators as first and second argument. May return duplicate
    tuples and tuples with times_chosen == 0.
    '''
    pprob = 0.0
    cprob = 0.0
    accumulated = 0
    iso_iter = population.__iter__()
    while count > 0:
        if accumulated > 0:
            yield (pop_next, accumulated)
            accumulated = 0
        pop_next, prob_next = next(iso_iter)
        pprob += prob_next
        # Beta mode
        while (pprob - cprob) * count / (1.0 - cprob) < 1.0:
            cprob += _beta_1_b(count) * (1.0 - cprob)
            while pprob < cprob:
                if accumulated > 0: 
                    yield (pop_next, accumulated)
                    accumulated = 0
                pop_next, prob_next = next(iso_iter)
                pprob += prob_next
            accumulated += 1
            count -= 1
            if count == 0: break
        if count == 0: break
        # Binomial mode
        nrtaken = _safe_binom(count, (pprob-cprob)/(1.0-cprob))
        accumulated += nrtaken
        count -= nrtaken
        cprob = pprob
    if accumulated > 0:
        yield (pop_next, accumulated)


class Sampler:
    def __init__(self, iso, ionic_current, accuracy, beta_bias):
        self.iso = iso.__iter__()
        self.accumulated_prob = 0.0
        self.chasing_prob = 0.0
        self.ionic_current = ionic_current
        self.accuracy = accuracy
        self.beta_bias = beta_bias
        self.accumulated = 0
        self.current_count = nan
    def advance(self):
        if self.ionic_current < 0:
            raise Exception("Ionic current < 0")
        if self.ionic_current <= 0:
            if self.accumulated > 0:
                print("LEFTOVER")
                self.current_count = self.accumulated
                self.accumulated = 0
                return True
            return False
        while self.chasing_prob >= self.accumulated_prob:
            self.mass, self.cconf_prob = next(self.iso)
            self.accumulated_prob += self.cconf_prob
        prob_diff = self.accumulated_prob - self.chasing_prob
        expected_mols = prob_diff * self.ionic_current
        rem_interval = self.accuracy - self.chasing_prob
        while True:
            if expected_mols / rem_interval < self.beta_bias:
                print("BETA")
                self.current_count = self.accumulated
                while self.ionic_current > 0 and self.chasing_prob < self.accumulated_prob:
                    self.chasing_prob += _beta_1_b(self.ionic_current) * rem_interval
                    self.ionic_current -= 1
                    self.current_count += 1
                    rem_interval = self.accuracy - self.chasing_prob
                if self.ionic_current > 0:
                    self.accumulated = 1
                    self.ionic_current -= 1
                def raisee():
                    raise Exception("current count < 0")
                return (True if self.current_count > 0 else raisee())
            else:
                print("BINOM")
                self.current_count = _safe_binom(self.ionic_current, prob_diff / rem_interval)
                self.ionic_current -= self.current_count
                self.current_count += self.accumulated
                self.accumulated = 0
                self.chasing_prob = self.accumulated_prob
                if self.current_count > 0:
                    return True
                self.mass, self.cconf_prob = next(self.iso)
                self.accumulated_prob += self.cconf_prob
                prob_diff = self.cconf_prob
                expected_mols = prob_diff * self.ionic_current
                rem_interval = self.accuracy - self.chasing_prob


    def current(self):
        return (self.mass, self.current_count)



def sample_isospec2(formula, count, precision):
    population = IsoSpecPy.IsoLayeredGenerator(formula, t_prob_hint = precision, reorder_marginals = False)
    S = Sampler(population, count, precision, 1.0)
    while S.advance():
        yield S.current()



class CIIC:
    def __init__(self, iso, no_confs, precision = 0.9999, beta_bias = 1.0):
        self.iso = iso.__iter__()
        self.confs_prob = 0.0
        self.chasing_prob = 0.0
        self.to_sample_left = no_confs
        self.precision = precision
        self.beta_bias = beta_bias
    def next(self):
        while True:
            if self.to_sample_left <= 0:
                return False
            if self.confs_prob < self.chasing_prob:
                # Beta was last
                self.current_count = 1
                self.to_sample_left -= 1
                self.current_conf, self.current_prob = next(self.iso)
                self.confs_prob += self.current_prob
                while self.confs_prob <= self.chasing_prob:
                    self.current_conf, self.current_prob = next(self.iso)
                    self.confs_prob += self.current_prob
                if self.to_sample_left <= 0:
                    assert self.current_count > 0
                    return True
                curr_conf_prob_left = self.confs_prob - self.chasing_prob
            else:
                # Binomial was last
                self.current_count = 0
                self.current_conf, self.current_prob = next(self.iso)
                self.confs_prob += self.current_prob
                curr_conf_prob_left = self.current_prob

            assert self.to_sample_left > 0
            assert self.chasing_prob < self.confs_prob

            prob_left_to_1 = self.precision - self.chasing_prob
            expected_confs = curr_conf_prob_left * self.to_sample_left / prob_left_to_1

            if self.beta_bias < 0.0:
                cond = choice([True, False])
                print("RAND")
            else:
                cond = expected_confs <= self.beta_bias
            if cond:
                # Beta mode: we keep making beta jumps until we leave the current configuration
                self.chasing_prob += _beta_1_b(self.to_sample_left) * prob_left_to_1
                while self.chasing_prob <= self.confs_prob:
                    self.current_count += 1
                    self.to_sample_left -= 1
                    if self.to_sample_left == 0:
                        return True
                    prob_left_to_1 = self.precision - self.chasing_prob
                    self.chasing_prob += _beta_1_b(self.to_sample_left) * prob_left_to_1
                if self.current_count > 0:
                    return True
            else:
                # Binomial mode: a single binomial step
                rbin = _safe_binom(self.to_sample_left, curr_conf_prob_left/prob_left_to_1)
                self.current_count += rbin
                self.to_sample_left -= rbin
                self.chasing_prob = self.confs_prob
                if self.current_count > 0:
                    return True



def sample_ciic(formula, count, precision):
    population = IsoSpecPy.IsoLayeredGenerator(formula, t_prob_hint = precision, reorder_marginals = False)
    S = CIIC(population, count, precision, -1.0)
    while S.next():
        print(S.confs_prob, S.chasing_prob)
        yield (S.current_conf, S.current_count)



from IsoSpecPy.Formulas import *
from scipy.stats import chisquare
import sys

if __name__ == '__main__':
    test_mol = surcose
    count = 10000000

    print("Starting...")
    X = sorted(x for x in IsoSpecPy.IsoThresholdGenerator(formula=test_mol, threshold=sys.float_info.min, reorder_marginals = False) if x[1] > 0)

    print("No configs: " + str(len(X)))

    Y = dict([(v[0], 0) for v in X])
    #print(Y)

    s = 0
    for x in sample_ciic(test_mol, count, 0.999999):
        print(x)
        Y[x[0]] = x[1]
        s += x[1]
    print("S:", s)
    assert s == count

    #print(X)
    #print(Y)

    X = [x[1]*count for x in sorted(X)]
    Y = [x[1] for x in sorted(Y.items())]

    #print(X)
    #print(Y)
    print(chisquare(Y, X))



# -*- coding: utf-8 -*-
#
#   Copyright (C) 2015-2020 Mateusz Łącki and Michał Startek.
#
#   This file is part of IsoSpec.
#
#   IsoSpec is free software: you can redistribute it and/or modify
#   it under the terms of the Simplified ("2-clause") BSD licence.
#
#   IsoSpec is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  
#
#   You should have received a copy of the Simplified BSD Licence
#   along with IsoSpec.  If not, see <https://opensource.org/licenses/BSD-2-Clause>.
#


'''
Bunch of deprecated functions for 1.0.X compatibility.
Avoid using them: there is a considerable overhead associated
with using the old interface... The backward compatibility module
is also very rudimentary, somewhat incomplete and not very well
tested too...

The current API is implemented in __init__.py, use that instead
'''

try:
    xrange
except NameError:
    xrange = range

import re
from .IsoSpecPy import IsoTotalProb, IsoThreshold

class IsoSpec():
    def __init__(
                    self,
                    _atomCounts,
                    _isotopeMasses,
                    _isotopeProbabilities,
                    _stopCondition,
                    tabSize = 1000,     # ignored
                    hashSize = 1000,    # ignored
                    step = 0.3,         # ignored
                    trim = True,        # True not supported yet, treated as False anyway
                    method = 'layered'
                ):

        isoargs = {
            "formula"           : None,
            "get_confs"         : True,
            "atomCounts"        : _atomCounts,
            "isotopeMasses"     : _isotopeMasses,
            "isotopeProbabilities"  : _isotopeProbabilities,
        }


        self.dimNumber                 = len(_atomCounts)
        self._isotopeNumbers           = [len(x) for x in _isotopeMasses]
        self.allIsotopeNumber         = sum(self._isotopeNumbers)
        self._atomCounts               = _atomCounts
        self._isotopeMasses            = _isotopeMasses
        self._isotopeProbabilities     = _isotopeProbabilities
        self._stopCondition            = _stopCondition

        from .__init__ import IsoThreshold

        try:
            algo = { 'layered' : lambda total_prob: IsoTotalProb(total_prob, **isoargs),
              'ordered' : lambda total_prob: IsoTotalProb(total_prob, **isoargs),
              'threshold_absolute' : lambda threshold: IsoThreshold(threshold, True, **isoargs),
              'threshold_relative' : lambda threshold: IsoThreshold(threshold, False, **isoargs),
              'layered_estimating' : lambda total_prob: IsoTotalProb(total_prob, **isoargs)
            }[method]
        except KeyError:
            raise Exception("Invalid ISO method")

        # Reference to iso needs to be held in this object: it will deallocate masses/lprobs/etc arrays on C++ side if we 
        # allow GC to collect it prematurely
        self.iso = algo(_stopCondition)

        self.masses = self.iso.masses
        self.lprobs = self.iso.lprobs
        self.probs  = self.iso.probs
        self.confs  = self.iso.confs
        self.size   = self.iso.size

        if method == 'ordered' and len(self.masses) > 0:
            L = sorted(zip(self.masses, self.lprobs, self.probs, self.confs), key = lambda x: -x[1])
            self.masses, self.lprobs, self.probs, self.confs = zip(*L)

    @staticmethod
    def IsoFromFormula(formula, cutoff, tabSize = 1000, hashSize = 1000, classId = None, method = 'threshold_relative', step = 0.25, trim = True):
        # It's much easier to just parse it in python than to use the C parsing function
        # and retrieve back into Python the relevant object sizes
        symbols = re.findall("\D+", formula)
        atom_counts = [int(x) for x in re.findall("\d+", formula)]

        if not len(symbols) == len(atom_counts):
            raise ValueError("Invalid formula")

        from .PeriodicTbl import symbol_to_masses, symbol_to_probs
        try:
            masses = tuple(symbol_to_masses[symbol] for symbol in symbols)
            probs = tuple(symbol_to_probs[symbol] for symbol in symbols)
        except KeyError:
            raise ValueError("Invalid formula")

        return IsoSpec(atom_counts, masses, probs, cutoff, tabSize, hashSize, step, trim, method)



    def __len__(self):
        return self.size

    def getConfsRaw(self):
        return (self.masses, self.lprobs, self.confs)

#    def get_conf_by_no(self, clist, idx):
#        idx *= self.allIsotopeNumber
#        ret = []
#        for ison in self._isotopeNumbers:
#            ret.append(tuple(clist[idx:idx+ison]))
#            idx += ison
#        return tuple(ret)


    def getConfs(self):
        masses, logProbs, isoCounts = self.getConfsRaw()
        rows_no = len(masses)
        masses  = list(masses)
        logProbs= list(logProbs)
        confs = []
        for i in xrange(rows_no):
            confs.append([x for sublist in isoCounts[i] for x in sublist])
        return masses, logProbs, confs


    def getConfsNumpy(self):
        import numpy as np
        masses, logProbs, configurations = self.getConfsRaw()
        rows_no = len(masses)
        if rows_no == 0:
            return np.array([])
        cols_no = len(configurations)//len(masses)
        masses  = np.array(list(masses))
        logProbs= np.array(list(logProbs))
        configurations = np.array(list(configurations)).reshape((rows_no,cols_no))
        return masses, logProbs, configurations


    def splitConf(self, l, offset = 0):
        conf = []
        idx = self.allIsotopeNumber * offset
        for i in xrange(self.dimNumber):
            conf.append(tuple(l[idx:idx+self._isotopeNumbers[i]]))
            idx += self._isotopeNumbers[i]
        return tuple(conf)

    def confStr(self, conf):
        return '\t'.join([' '.join([str(x) for x in y]) for y in conf])

    def printConfs(self):
        masses, logProbs, isoCounts = self.getConfsRaw()
        confs = []
        step = sum(self._isotopeNumbers)
        for i in xrange(len(masses)):
            confs.append((masses[i], logProbs[i], self.splitConf(isoCounts, i)))

        for conf in confs:
            print(("Mass = {0}\t and log-prob = {1} and prob = {2}\t and configuration"\
                  "=\t{3}").format(conf[0], conf[1], math.exp(conf[1]), self.confStr(conf[2])))



class IsoPlot(dict):
    def __init__(self, iso, bin_w):
        self.iso = iso
        self.bin_w = bin_w
        masses, logProbs, _isoCounts = iso.getConfsRaw()
        dd = defaultdict(Summator)
        for i in xrange(len(masses)):
            dd[float(int(masses[i]/bin_w))*bin_w].add(math.exp(logProbs[i]))
        for key, val in dd.items():
            self[key] = val.get()


def IsoSpecify( formula,
                cutoff,
                method= 'layered',
                output_format = 'numpy_arrays',
                trim  = True,
                _step = 0.25,
                _trim = True,
                _tabSize  = 1000,
                _hashSize = 1000    ):
    """
    Call IsoSpec on a formula with a given cutoff.

    This function wraps around the IsoSpec class.

    Parameters
    ----------
    formula : char
        a string of a form '< Element Tag 1 >< Count 1 > ... ',
        e.g. 'C100H202'. Using IUPAC conventions to name elements.

    cutoff : float
        The cutoff value. See description of the method argument.

    method : char
        Can take one of the following values: 'layered',
        'layered_estimating', 'threshold_absolute',
        'threshold_relative', 'ordered'.

        The threshold versions of the algorithm rely on user
        providing a precise lower bound on the reported peak heights.
        This can be specified in absolute terms ('threshold_absolute'),
        i.e. in terms of the limiting probability of the isotopologue,
        or as a percentage of the heighest peak ('threshold_relative').

        The layered versions of the algorithm rely on calculating
        consecutive values of peak thresholds on flight.
        The ultimate goal is to reach a peak probability that assures
        that the sum of probabilities of the more probable isotopologues
        exceeds the provided cutoff value.
        The sequence of consecutive thresholds can be generated in
        two ways. The default way, 'layered_estimating', estimates
        the threshold to joint probability function by a progressive
        linear spline, check Anal Chem. 2017 Mar 21;89(6):3272-3277.
        doi: 10.1021/acs.analchem.6b01459. Epub 2017 Mar 8.
        The other way, 'layered', estimates a threshold as a 30%%
        quantile of the probabilities gathered in the fringe set, i.e.
        isotopologues that are direct neighbours of the previously
        accepted layer. Finally, choosing the 'ordered' version will
        provide a loglinear version of the algorithm that relies on
        the priority queue. This version automatically sorts
        the configurations by their probability.

    trim
        while using a layered method, should one discard superfluously
        obtained isotopologues, i.e. such that without them the set of
        reported isotopologues already is an optimal p-set.

    output_format
        Should the output be presented as lists ('lists'),
        or as numpy arrays ('numpy_arrays').

    Returns
    -------
    masses
        masses of isotopologues, either a list or a numpy array.

    logProbs
        logarithms of probabilities (theoretical heights) of isotopologues,
        either a list or a numpy array.

    confs
        counts of isotopologues (extended chemical formulas that
        include counts of isotopes of elements)
        """


    assert output_format in ('lists', 'numpy_arrays'), "Wrong value of output_format. Should be either 'lists' or 'numpy_arrays'."

    assert method in ('layered', 'ordered', 'threshold_absolute', 'threshold_relative', 'layered_estimating'), "Wrong value of method. Should be among 'layered', 'ordered', 'threshold_absolute', 'threshold_relative', or 'layered_estimating'."

    assert isinstance(cutoff, float), "Provided cut off ain't a float."

    assert isinstance(formula, str), "Provided formula off ain't a string."

    iso = IsoSpec.IsoFromFormula(   formula,
                                    cutoff,
                                    tabSize = 1000,
                                    hashSize = 1000,
                                    classId = None,
                                    method = method,
                                    step = 0.25,
                                    trim = trim )

    if output_format == 'lists':
        masses, logProbs, confs = iso.getConfs()
    else:
        masses, logProbs, confs = iso.getConfsNumpy()

    # print 'Rev Startek is a silly old chump and his mother dresses up silly.'
    return masses, logProbs, confs


